//---------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <time.h>


float **cria_matriz_float(int nlinhas, int ncolunas);
int *cria_vetor(int tam);
void le_arq_producao(char nomearq[], int nlinhas, int ncolunas, float **producao);
void libera_vetor(int *vetor);
void libera_matriz_float(float **matriz, int nlinhas);
void constroi_solucao_inicial(int *s, int tam, int nopcoes);
float calcula_fo(int *s,
                 float **producao,
                 float **vpl,
                 int nlinhas,
                 float meta,
                 int penalidade_menos,
                 int penalidade_mais);

float SA(int *s,
         float **producao,
         float **vpl,
         int nlinhas,
         int ncolunas,
         float meta,
         int penalidade_menos,
         int penalidade_mais,
         float temperatura_inicial,
         float temperatura_final,
         int SAmax,
         float alpha);

float temp_inicial(int nlinhas,
                   int *s,
                   float **vpl,
                   float beta,
                   float gamma,
                   int SAmax);

float calcula_delta(int nlinhas, int *s, float **vpl, int i, int j);

float randomico(float min, float max);
void atualiza_vetor(int *recebe, int *doa, int n);

float descida(int *s,
              float **producao,
              float **vpl,
              int nlinhas,
              int ncolunas,
              float meta,
              int penalidade_menos,
              int penalidade_mais);

float melhor_vizinho(int *s,
                     float **producao,
                     float **vpl,
                     int nlinhas,
                     int ncolunas,
                     float meta,
                     int penalidade_menos,
                     int penalidade_mais,
                     float fo,
                     int melhor_i,
                     int *melhor_j);


// ******************* Iniciando ***************************

int main(int argc, char* argv[])
{
  float **producao,    // = int producao[7][16]
        **vpl,
        fo;
  float temperatura_inicial, temperatura_final;
  int nlinhas, ncolunas;
  int *s;
  clock_t inicioCPU, fimCPU;

  inicioCPU = fimCPU = clock();
  nlinhas = 7;
  ncolunas = 16;

  int semente = time(NULL);
  srand(semente);
  s = cria_vetor(nlinhas);
  producao = cria_matriz_float(nlinhas, ncolunas);
  vpl = cria_matriz_float(nlinhas, ncolunas);
  le_arq_producao("producao.txt",nlinhas, ncolunas,producao);
  le_arq_producao("vpl.txt",nlinhas, ncolunas,vpl);
  constroi_solucao_inicial(s,nlinhas,ncolunas);

  //  Solucao tima e do AG do Joaquim
  /*
  s[0] = 6;
  s[1] = 14;
  s[2] = 6;
  s[3] = 2;
  s[4] = 7; //2;
  s[5] = 13; //9;
  s[6] = 4;
  // soluo tima = fo=2180
  */

  fo = calcula_fo(s,producao,vpl,nlinhas,90,3,15);
  printf("fo inicial =%f \n\n",fo);
  // ----------- Temperatura Inicial --------------------------
  temperatura_inicial = temp_inicial(nlinhas,s,vpl,1.1,0.95,500);
  printf("Temperatura inicial =%f \n\n",temperatura_inicial);
  temperatura_final = 0.01;
  // -------------------------Simulated Annealing--------------
  fo = SA(s,producao,vpl,nlinhas,ncolunas,90,3,15,temperatura_inicial,
          temperatura_final,500,0.998);
  fimCPU = clock();
  // ------------------------- Mtodo da descida ---------------
     fo = descida(s, producao, vpl,nlinhas, ncolunas,90,3,15);
  // -------funo objetivo----------
  printf("OPCAO SELECIONADA EM CADA MINA\n");
  for (int i=0; i<nlinhas; i++)
    printf("Mina %d=%d \n",i+1,s[i]);
  printf("Semente = %d \t Tempo de execucao = %8.2f \n\n", semente,
                          (fimCPU - inicioCPU)/CLOCKS_PER_SEC);
  printf("fo =%f \n",fo);
  libera_vetor(s);
  libera_matriz_float(producao,nlinhas);
  libera_matriz_float(vpl,nlinhas);
  getchar();
  return 0;
}
//********************* terminando ********************************

/* cria memoria para um vetor de tam posicoes */
int *cria_vetor(int tam)
{
  int *vetor;

  vetor = (int *) malloc(tam*sizeof(int));
  if (!vetor){
  	printf("Falta memoria para alocar o vetor de ponteiros");
    exit(1);
  }
  return vetor;
}

void le_arq_producao(char nomearq[], int nlinhas, int ncolunas,
                     float **producao)
{
  FILE *arquivo;
  int i, j;
  float valor;

  i = 0;
  arquivo = fopen(nomearq,"r");
  if (!arquivo) {
     printf("O Arquivo %s nao pode ser aberto.\n", nomearq);
     getchar();
     exit(1);
  }
  while (i < nlinhas){
    j = 0;
    while (j < ncolunas){
      fscanf(arquivo, "%f",&valor);
      producao[i][j] = valor;
      j++;
    }
    i++;
  }
  fclose(arquivo);
}

/* Cria matriz de ponteiros para inteiros com nlinhas e ncolunas */
float **cria_matriz_float(int nlinhas, int ncolunas)
{
  register int i;
  float **matriz;

  matriz = (float **) malloc(nlinhas*sizeof(float *));
  if (!matriz) {
        printf("Falta memoria para alocar a matriz de ponteiros\n");
        exit(1);
  }
  for (i=0; i < nlinhas; i++) {
    matriz[i] = (float *) malloc(ncolunas*sizeof(float));
    if (!matriz[i]){
      printf("Falta memoria para alocar a matriz de ponteiros.\n");
      exit(1);
    }
  }
  return matriz;
}


/* libera memoria de um vetor */
void libera_vetor(int *vetor)
{
  free(vetor);
}

void libera_matriz_float(float **matriz, int nlinhas)
{
  register int i;

  for (i=nlinhas-1; i >= 0; i--)
    free((float *) matriz[i]);
  free((float *) matriz);
}


/* constri soluo inicial */
void constroi_solucao_inicial(int *s, int tam, int nopcoes)
{
  int i;

  for (i=0; i<tam; i++){
    s[i] = random(nopcoes);
  }
}

/* calcula funo objetivo */
float calcula_fo(int *s,
                 float **producao,
                 float **vpl,
                 int nlinhas,
                 float meta,
                 int penalidade_menos,
                 int penalidade_mais)
{
  int i,j;
  float fo, soma;
  soma = 0;

  /* Calcular vpl */
  for (i = 0; i<nlinhas; i++){
    soma = soma + vpl[i][s[i]];
  }
  fo = soma;

  /* Calcular penalidade de producao */
  soma = 0;
  for (i = 0; i<nlinhas; i++){
    soma = soma + producao[i][s[i]];
  }
  if (soma < meta)
    soma = (meta - soma)*penalidade_menos;
  if (soma > meta)
    soma = (soma - meta)*penalidade_mais;

  fo = fo - soma;
  return fo;
}

/* mtodo simulated Annealing */
float SA(int *s,
         float **producao,
         float **vpl,
         int nlinhas,
         int ncolunas,
         float meta,
         int penalidade_menos,
         int penalidade_mais,
         float temperatura_inicial,
         float temperatura_final,
         int SAmax,
         float alpha)
{
  int iterSA, i;
  int *sstar, *slinha;
  float fo, temperatura, folinha, fostar, delta, x;

  sstar = cria_vetor(nlinhas);
  slinha = cria_vetor(nlinhas);
  temperatura = temperatura_inicial;
  iterSA = 0;
  atualiza_vetor(sstar,s,nlinhas);
  fo = fostar = calcula_fo(s,producao,vpl,nlinhas,meta,penalidade_menos,
                           penalidade_mais);
  while (temperatura > temperatura_final) {
    while (iterSA < SAmax) {
      iterSA++;
      // Gere um vizinho qualquer
      // ---------- Faz o movimento -------------
         atualiza_vetor(slinha, s, nlinhas);
         i = random(nlinhas);
         if (slinha [i] < ncolunas - 1)
           slinha [i] = slinha [i] + 1;
          else
           slinha [i] = 0;

      // ---------- fim do movimento ------------

      fo = calcula_fo(s,producao,vpl,nlinhas,meta,penalidade_menos,
                      penalidade_mais);
      folinha = calcula_fo(slinha,producao,vpl,nlinhas,meta,penalidade_menos,
                           penalidade_mais);
      float delta = folinha - fo;
      if (delta > 0) {
        atualiza_vetor(s,slinha,nlinhas);
        fo = folinha;
        if (fo > fostar){
          fostar = fo;
          atualiza_vetor(sstar, s, nlinhas);
         }
      }
      else {
        x = randomico(0,1);
        if (x < exp(delta/temperatura)){
          atualiza_vetor(s, slinha, nlinhas);
          fo = folinha;
        }
        // Desfazer o movimento
      }
    }
    temperatura = alpha * temperatura;
    iterSA = 0;
  }
  atualiza_vetor(s,sstar,nlinhas);
  fo = fostar;
  libera_vetor(sstar);
  libera_vetor(slinha);
  return fo;
}

/* Gera numero aleatorio entre min e max */
 float randomico(float min, float max)
{
  if (min == max) return min;
  return ((float) (rand()%10000/10000.0)*(max-min) + min);
}

/* atualiza a melhor solucao */
void atualiza_vetor(int *recebe, int *doa, int n)
{
   for (int j=0; j < n; j++) recebe[j] = doa[j];
}

/* calcula temperatura inicial */
float temp_inicial(int nlinhas,
                   int *s,
                   float **vpl,
                   float beta,
                   float gamma,
                   int SAmax)
{
  int i, j, aux, iterT, aceitos;
  float temperatura, delta, fo;
  bool continua;

  temperatura = 10;
  continua = true;
  while (continua){
    aceitos = 0;
    iterT = 0;
    while (iterT < SAmax){
      iterT++;
      // Gere um vizinho qualquer
      i = random(nlinhas);
      do{
        j = random(nlinhas);
      } while (j == i);
      float delta1 = calcula_delta(nlinhas, s, vpl, i, j);
      // Faz o movimento
      aux = s[j];
      s[j] = s[i];
      s[i] = aux;
      float delta2 = calcula_delta(nlinhas, s, vpl, i, j);
      delta = - delta1 + delta2;
      if (delta > 0){
         aceitos++;
      }
      else{
        float x = randomico(0,1);
        if (x < exp(delta/temperatura)){
           aceitos++;
        }
      }
      // Desfaz o movimento
      aux = s[j];
      s[j] = s[i];
      s[i] = aux;
    }
    if (aceitos < gamma * SAmax){
      temperatura = beta * temperatura;
    }
    else
      continua = false;
  }
  return temperatura;
}

/* calcula delta */
float calcula_delta(int nlinhas, int *s, float **vpl, int i, int j)
{
  int i_antes, i_depois, j_antes, j_depois;

  i_antes = i - 1;
  i_depois = i + 1;
  j_antes = j - 1;
  j_depois = j + 1;
  if (i == 0) i_antes = nlinhas-1;
  if (i == nlinhas-1) i_depois = 0;
  if (j == 0) j_antes = nlinhas-1;
  if (j == nlinhas-1) j_depois = 0;

 return  vpl[i_antes][s[i]] + vpl[i][s[i_depois]] + vpl[j_antes][s[j]] +
         vpl[j][s[j_depois]];
}

/*calcula melhor vizinho*/
float melhor_vizinho(int *s,
                     float **producao,
                     float **vpl,
                     int nlinhas,
                     int ncolunas,
                     float meta,
                     int penalidade_menos,
                     int penalidade_mais,
                     float fo,
                     int melhor_i,
                     int *melhor_j)

{
  int aux;
  float fo_melhor_viz = fo;

  float fo_viz;
  aux = s[melhor_i];
  for(int j=0 ; j < ncolunas ; j++){
     // Fazer o movimento
     if (j != s[melhor_i]){
      s[melhor_i] = j;
      fo_viz = calcula_fo(s,producao,vpl,nlinhas,meta,penalidade_menos,
                          penalidade_mais);

      // Armazenar o melhor movimento (melhor troca)
      if(fo_viz > fo_melhor_viz){
        *melhor_j = j;
        fo_melhor_viz = fo_viz;
      }
     }
  }
   // Desfazer o movimento
      s[melhor_i] = aux;
return fo_melhor_viz;
}

/* mtodo da descida */
float descida(int *s,
              float **producao,
              float **vpl,
              int nlinhas,
              int ncolunas,
              float meta,
              int penalidade_menos,
              int penalidade_mais)
{
  int aux, melhor_j, melhor_i;
  float fo_viz, fo;
  bool melhorou;

  do{
    melhorou = false;
    for (int i=0; i<nlinhas; i++){
      fo = fo_viz = calcula_fo(s,producao,vpl,nlinhas,meta,penalidade_menos,
                               penalidade_mais);
      fo_viz = melhor_vizinho(s,producao,vpl,nlinhas,ncolunas,meta,penalidade_menos,
                          penalidade_mais,fo,i,&melhor_j);
       if (fo_viz > fo){
          s[i] = melhor_j;
          fo = fo_viz;
          melhorou = true;
       }
    }
  }while (melhorou);
  return fo;
}





